from dataclasses import dataclass, replace
from typing import Any, Optional, NamedTuple
import pandas as pd, numpy as np

from .data_structures import HistogramData, HistogramFactory
from .config import RheoscaleConfig
from .errors import RheoscaleError
from .schemas import POSITION_COLUMNS


from .policy.check_input_cofig import check_and_update_config
from .policy.create_running_config import create_config_w_data
from .policy.rheoscores import calculate_rheoscores


from .io import load_data, validate_columns, write_outputs


class RheoscaleRunner:
    """
    High-level orchestrator for RheoScale.

    Responsibilities:
    - validate inputs
    - apply config-driven policy
    - coordinate computation
    - handle errors and IO
    """

    def __init__(self, config: RheoscaleConfig, DMS_data: Optional[pd.DataFrame] = None):
        # starting data
        self.user_config: RheoscaleConfig = config
        self.DMS_data: pd.DataFrame = self._load_data(DMS_data) 
        
        
        # added to while running
        self.running_config: RheoscaleConfig = None
        
        self.histogram_factory: HistogramFactory = None
        self.rheoscale_position_data: pd.DataFrame = None


    def _load_data(self, dms_data):
        
        if self.user_config.input_file_name is not None and dms_data is None:
            data_from_dms = load_data(self.user_config)
        elif self.user_config.input_file_name is None and dms_data is not None:
            data_from_dms =dms_data
        else:
            raise ValueError('Either a input file must be entered within the configeration XOR a df must be added during this call')   

        validate_columns(data_from_dms, self.user_config)
        
        data_from_dms = data_from_dms.dropna(how="all")

        if self.user_config.log_scale:
            data_from_dms = self.transform_raw_data(data_from_dms)

        if self.user_config.error_val is not None:
            data_from_dms[self.user_config.columns['error']]  = self.user_config.error_val

        

        return data_from_dms  
    


    def run(self):
        #try:
            
            #checking or creating inputs
            update = self.test_accuracy_of_input_config()
            
            #create a running config
            self.create_analysis_config(update)
            self.histogram_factory = self.calculate_bins_and_weight()
            
            #doing rheoscale analysis
            self.rheoscale_position_data = self.rheoscale_analysis()
            
            #writing outputs
            if self.running_config.output_dir != '':
                self.output_data()

            return self.rheoscale_position_data


    def test_accuracy_of_input_config(self):
        
        return check_and_update_config(self.user_config, self.DMS_data)
        

    def transform_raw_data(self, DMS_data):
        DMS_data[self.user_config.columns['error']] = 0.434*((DMS_data[self.user_config.columns['error']])/(DMS_data[self.user_config.columns['value']]))
                                                        
        DMS_data[self.user_config.columns['value']] = np.log10(DMS_data[self.user_config.columns['value']])

        
        return DMS_data
        

    def create_analysis_config(self, update:dict):
        if self.running_config is not None: 
            raise ValueError('This should only be run once')
        
        self.running_config = replace(self.user_config,
                   number_of_positions= update['number_of_positions'],
                   WT_val= update['WT_val'],
                   WT_error= update['WT_error'],
                   min_val= update['min_val'],
                   max_val= update['max_val'],
                   error_val= update['error_val'],
                   number_of_bins= update['number_of_bins'],
                   neutral_binsize= update['neutral_binsize'],
                   _true_min  = update['_true_min'],
                   _true_max = update['_true_max']
        )
        self.running_config._validate_and_make_output()
 
    def calculate_bins_and_weight(self):
        bins_size = np.abs(self.running_config.min_val-self.running_config.max_val)/self.running_config.number_of_bins
        if self.running_config.neutral_binsize == 0.0 or self.running_config.neutral_binsize == 0:
            self.running_config = replace(self.running_config, neutral_binsize=bins_size*2)

        #add one for the final edge
        if self.running_config.dead_extremum == 'Min':
            dead = np.linspace(start=self.running_config._true_min, stop=bins_size+self.running_config.min_val, num=1)
            remaining= np.linspace(start=bins_size+self.running_config.min_val, stop=self.running_config.max_val, num=self.running_config.number_of_bins)
            if dead != self.running_config.min_val:
                bin_edges = np.concatenate([dead, remaining])
            else: 
                bin_edges = remaining
        else: 
             dead = np.linspace(start=self.running_config._true_max, stop=bins_size+self.running_config.max_val, num=1)
             remaining= np.linspace(start=bins_size+self.running_config.min_val, stop=self.running_config.max_val, num=self.running_config.number_of_bins)
             bottom = np.linspace(start=remaining[0]-bins_size, stop=remaining[0], num=1)
             
             if dead != self.running_config.max_val:
                 bin_edges = np.concatenate([bottom, remaining, dead])
             else: 
                 bin_edges = np.concatenate([bottom, remaining])
        wt_bin = np.digitize(self.running_config.WT_val, bin_edges)-1

        if self.running_config.dead_extremum == "Min":
            
            dead_bin = 0
        else:
            
            dead_bin= np.array(range(self.running_config.number_of_bins)).max()
        
        weights = np.full(self.running_config.number_of_bins, 3)
        
        for weight in range(self.running_config.number_of_bins):
            if abs(weight - dead_bin) == 1:
                weights[weight] = 2
            if abs(weight - wt_bin) == 1:
                weights[weight] = 2
            if weight == dead_bin:
                weights[weight] = 1
            if weight == wt_bin:
                weights[weight] = 1

        hist_fact = HistogramFactory(bin_edges=bin_edges, weights=weights)

        return  hist_fact

    def rheoscale_analysis(self):
        
        
        return calculate_rheoscores(self.running_config, self.DMS_data, self.histogram_factory)
        

    def output_data(self):
        write_outputs(self.running_config, self.rheoscale_position_data)

        



import numpy as np
from typing import NamedTuple, Union
from dataclasses import dataclass, fields

class HistogramData(NamedTuple):
    counts: np.ndarray
    bin_edges: np.ndarray
    weights: np.ndarray = None

    def __add__(self, other):
        if not isinstance(other, HistogramData):
            return NotImplemented

        # Bin compatibility check
        if not np.array_equal(self.bin_edges, other.bin_edges):
            raise ValueError("Cannot add histograms with different bin edges")

        # Add counts
        new_counts = self.counts + other.counts

        # Handle weights
        if self.weights is None and other.weights is None:
            new_weights = None
        elif self.weights is None:
            new_weights = other.weights.copy()
        elif other.weights is None:
            new_weights = self.weights.copy()
        else:
            new_weights = self.weights + other.weights

        return HistogramData(
            counts=new_counts,
            bin_edges=self.bin_edges,
            weights=new_weights
        )     
        


class HistogramFactory:
    def __init__(self, bin_edges, weights=None):
        self.bin_edges = bin_edges
        self.weights = weights
    
    def create_hist_data(self, data):
        counts, _ = np.histogram(data, bins=self.bin_edges)
        return HistogramData(counts, self.bin_edges, self.weights)


@dataclass(slots=True)
class RheoScores:
    position: int
    num_of_variants: int
    histogram: HistogramData 
    neutral_score: float=None
    toggle_score: float=None
    rheostat_score: float=None
    weighted_rheostat_score: float=None
    enhancing_score: float=None
    binary: bool=None
    average: float = None
    st_dev: float= None
    assignment: str = None

    def __setattr__(self, name, value):
        if name == 'position' or name == 'num_of_variants' or name == 'histogram':
            pass
        elif name == 'assignment':
            if value not in ['neutral', 'toggle', 'rheostat', 'enhancing', 'adverse', 'WT/inactive','moderate', None, 'unclassified']:
                raise TypeError(f'{name} must be a name of a position type')
        elif name in {"binary"}:
            if not isinstance(value, Union[None, bool]):
                raise TypeError(f"{name} must be bool")
        elif not isinstance(value, Union[None, int, float, HistogramData]):
            raise TypeError(f"{name} must be numeric")
        
        
        super().__setattr__(name, value)

    


# schemas.py
POSITION_COLUMNS = [
    'position',
    'num_of_variants',
    'histogram',
    'enhancing_score',
    'neutral_score',
    'rheostat_score',
    'weighted_rheostat_score',
    'toggle_score',
    'binary',
    'average',
    'st_dev',
    'assignment',
]




import argparse
from pathlib import Path
from rheoscale.config import RheoscaleConfig  
from rheoscale.rheoscale_runner import RheoscaleRunner


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="rheoscale",
        description="Run RheoScale analysis with configurable parameters"
    )

    # --- Core ---
    parser.add_argument(
        "protein_name",
        help="Name of the protein being analyzed"
    )

    parser.add_argument(
        "-i", "--input-file",
        dest="input_file_name",
        type=str,
        help="Path to input CSV file"
    )

    parser.add_argument(
        "-n", "--num-positions",
        dest="number_of_positions",
        type=int,
        help="Number of positions to analyze"
    )

    # --- WT options ---
    parser.add_argument("--wt-val", type=float, help="WT value")
    parser.add_argument("--wt-error", type=float, help="WT error")
    parser.add_argument("--wt-name", type=str, default="WT")

    # --- Value ranges ---
    parser.add_argument("--min-val", type=float)
    parser.add_argument("--max-val", type=float)
    parser.add_argument("--error-val", type=float)

    # --- Binning ---
    parser.add_argument("--bins", dest="number_of_bins", type=int)
    parser.add_argument(
        "--dead-extremum",
        choices=["Min", "Max"],
        default="Min"
    )
    parser.add_argument("--neutral-binsize", type=float)

    # --- Flags ---
    parser.add_argument("--log-scale", action="store_true")
    parser.add_argument("--even-bins", action="store_true")
    parser.add_argument("--output-histograms", action="store_true")

    # --- Output ---
    parser.add_argument(
        "-o", "--output-dir",
        default="Rheoscale_analysis",
        help="Output directory"
    )

    # --- Config I/O ---
    parser.add_argument(
        "--save-config",
        type=Path,
        help="Save config to JSON file and exit"
    )

    parser.add_argument(
        "--load-config",
        type=Path,
        help="Load config from JSON file"
    )

    return parser


def main():
    parser = build_parser()
    args = parser.parse_args()

    # ---- Load from JSON if requested ----
    if args.load_config:
        config = RheoscaleConfig.from_json(args.load_config)
    else:
        config = RheoscaleConfig(
            protein_name=args.protein_name,
            input_file_name=args.input_file_name,
            number_of_positions=args.number_of_positions,
            log_scale=args.log_scale,
            WT_val=args.wt_val,
            WT_error=args.wt_error,
            WT_name=args.wt_name,
            min_val=args.min_val,
            max_val=args.max_val,
            error_val=args.error_val,
            number_of_bins=args.number_of_bins,
            dead_extremum=args.dead_extremum,
            neutral_binsize=args.neutral_binsize,
            output_dir=args.output_dir,
            output_histogram_plots=args.output_histograms,
            even_bins=args.even_bins,
        )

    # ---- Save config if requested ----
    if args.save_config:
        config.to_json(args.save_config)
        print(f"Config saved to {args.save_config}")
        return

    # ---- Run validations / setup ----
    config._validate_and_make_output()

    # ---- Hand off to analysis ----
    print("Running RheoScale with config:")
    print(config)
    if config.input_file_name is None:
        raise ValueError('For CLI you must have a file that you can upload to the config must have --input_file_name')
    RheoscaleRunner(config)


if __name__ == "__main__":
    main()



import pandas as pd
import numpy as np

from .data_structures import HistogramData
import matplotlib
import sys
if "debugpy" in sys.modules:
    matplotlib.use("Agg")

import matplotlib.pyplot as plt

def plot_all_positions(positions: list, hist_list: list, dead_extremum, WT_value,dead_value,save_path, neutral_bin_size, prefix='', all_pos=False, is_even_bins=False):
    master_hist_data = hist_list[0]
    for i in range(len(positions)):
        name = prefix+'_pos_'+str(positions[i])
        if all_pos:
             make_tuning_plot_one_pos(hist_list[i], dead_extremum, WT_value,dead_value, neutral_bin_size, save_path, tle=name, is_even_bins=is_even_bins)
        if i!=0:
             master_hist_data+= hist_list[i]
    all_title = prefix+'_all_positions'
    make_tuning_plot_one_pos(master_hist_data, dead_extremum, WT_value,dead_value, neutral_bin_size, save_path, all_title, is_all=True, is_even_bins=is_even_bins)

    
def make_tuning_plot_one_pos(hist_data: HistogramData , dead_extremum, WT_value, dead_value, neutral_bin_size,path,     
    tle: str = "Histogram",
    is_all: bool = False,
    is_even_bins: bool = False,
    xlabel: str = "Value",
    ylabel: str = "Count",
    log_y: bool = True,
    label_precision: int = 2):
        
        counts = hist_data.counts
        bin_edges = hist_data.bin_edges

        # ---- sanity check ----
        if len(bin_edges) != len(counts) + 1:
            raise ValueError("bin_edges must be one element longer than counts")
        if is_even_bins:
             if dead_extremum == 'Min':
                  bin_edges[0] = dead_value
             else:
                  bin_edges[-1] = dead_value
        
        # bin_widths = np.diff(bin_edges)
        # bin_centers = bin_edges[:-1] + bin_widths / 2

        bin_centers = 0.5 * (bin_edges[:-1] + bin_edges[1:])
        bin_widths = np.diff(bin_edges)

        
        fmt = f"{{:.{label_precision}f}}"
        bin_labels = [
            f"{fmt.format(left)}–{fmt.format(right)}"
            for left, right in zip(bin_edges[:-1], bin_edges[1:])
        ]

        # ---- create figure & axes ----
        fig, ax = plt.subplots()

        # ---- plot ----
        ax.bar(
            bin_centers,
            counts,
            width=bin_widths*0.85,
            align="center",
            color='black'
        )
        
        

        WT_index =  np.digitize(WT_value, bin_edges) - 1
        WT_bin_center = bin_centers[WT_index]

        ax.axvline(dead_value, color='red')
        ax.axvspan(WT_value-(neutral_bin_size/2), WT_value+(neutral_bin_size/2), alpha=.55, color='green')

        # ---- axes formatting ----
        if not is_all:
            ax.set_ylim(0.001, 21)
            ax.set_yticks([i for i in range(5,21, 5)])
        else:
            
            ax.set_yscale('log')
            ax.set_ylim(1)
        ax.set_xticks(bin_centers)
        ax.set_xticklabels(bin_labels, rotation=45, ha="right")

        ax.set_xlabel(xlabel)
        ax.set_title(tle)

        if log_y:
            pass
            #ax.set_yscale("log")

        fig.tight_layout()
        plt.savefig(rf'{path}\{tle}.png')
        print(fr'#####saved to {path}\{tle}.png')
        plt.close()

        




class RheoscaleError(Exception):
    """Base error for RheoScale."""

class ConfigError(RheoscaleError):
    pass

class DataError(RheoscaleError):
    pass

class ComputationError(RheoscaleError):
    pass

class OutputError(RheoscaleError):
    pass



from dataclasses import dataclass, field, replace, fields, asdict
from typing import Optional, Dict, Literal, Union, ClassVar,  get_origin, get_args, TypeAlias
from types import MappingProxyType
import os, pandas as pd
from pathlib import Path
#interal keys
Number = Union[int, float]
NumericFieldDict: TypeAlias = dict[str,None | Number]



ColumnKey = Literal[
    "position",
    "substitution",
    "value",
    "error"

]

class FixedKeysDict(dict):
    def __setitem__(self, key, value):
        if key not in self:
            raise KeyError(f"Key '{key}' cannot be added")
        super().__setitem__(key, value)

@dataclass (frozen=True)
class RheoscaleConfig:
    protein_name: str = field(
        metadata={"label": "Protein name"}
    )
    input_file_name: str = field(default=None, metadata={'label': 'Input File Path'})
    number_of_positions: int= field(default=None, metadata={'label': 'Number of positions'})
    log_scale: bool = field(default=False, metadata={'label': 'Convert values to log scale'})
    WT_val: float = field(default=None, metadata={'label': 'Measured value of WT'})
    WT_error: float= field(default=None, metadata={'label': 'Measured value of WT error'}) 
    WT_name: str = field(default=None, metadata={'label': 'Name of WT in position coloumn in data'})
    
    min_val: float = field(default=None, metadata={'label': 'Minimum measurable value'})
    max_val: float = field(default=None, metadata={'label': 'Maximum measurable value'}) 
    error_val: float = field(default=None, metadata={'label': 'Error override'}) 
    number_of_bins: int = field(
        default=None,
        metadata={
            "label": "Number of bins",
            "min": 2,
            "max": 20
        })
    dead_extremum: Literal["Min", "Max"]= field(default="Min", metadata={'label': 'Dead Extremum'})
    neutral_binsize: float = field(default=None, metadata={'label': 'Size of the neutral bin'}) 
    output_dir: str =field(default='Rheoscale_analysis', metadata={'label': 'Name of output dir that is added after the protein name'}) 
    
    output_histogram_plots: bool = field(default=False, metadata={'label': 'Output all the histogram plots of each position'}) 
    even_bins: bool = field(default=True, metadata={'label': 'Even bins (Even bins if True, if false dead bin will be as wide as measured values)'}) 
    _true_min: float= None
    _true_max: float= None
    '''
    potential bug is having WT val and error val at the same time'''

    # Threshold defaults
    enhancing_threshold: ClassVar[float]= 0.8
    neutral_threshold: ClassVar[float]= 0.7
    rheostat_threshold: ClassVar[float]= 0.5
    toggle_threshold: ClassVar[float]= 0.64
    
    columns: Dict[ColumnKey, str] = field(
        default_factory=lambda: MappingProxyType({
            "position": "Position",
            "substitution": 'Substitution',
            "value": 'Value',
            "error": 'Error'
        }), metadata= {'label': 'Name of Columns'})
    



    def __post_init__(self):
        self._validate_name()
        self._validate_input_file()
        self._validate_thresholds()
        self._validate_bins()
        self._validate_num_pos()
        
        self._validate_WT()
    def _validate_name(self):
        if not isinstance(self.protein_name, str):
            raise ValueError('The input name MUST be a string!')

    def _validate_WT(self):
        if self.WT_val is None and self.WT_error is None:
            pass
        elif self.WT_error is not None and self.WT_val is not None:
            pass
        elif self.WT_val is not None and self.WT_name is None:
            pass
        elif self.WT_val is None and self.WT_name is not None:
            pass
        else:
            raise ValueError('if the WT_val AND WT_error must be entered together') 

    def _validate_input_file(self):
        if self.input_file_name is not None:
            is_path = os.path.exists(self.input_file_name)
            if is_path and self.input_file_name.endswith('.csv'):
                pass
            else:
                raise FileNotFoundError(f'the path {self.input_file_name} was not found')
    

    def _validate_bins(self):
        if self.number_of_bins is not None:
            if self.number_of_bins > 20 or self.number_of_bins <=1:
                raise ValueError(f"you entered a number_of_bins = {self.number_of_bins}, you must have at least 2 bins and no more than 20")
    

    def _validate_thresholds(self):
        for name in (
            "enhancing_threshold",
            "neutral_threshold",
            "rheostat_threshold",
            "toggle_threshold",
        ):
            value = getattr(self, name)
            if not (0.0 <= value <= 1.0):
                raise ValueError(
                    f"{name} must be between 0 and 1 inclusive, got {value}"
                )
            
    def _validate_num_pos(self):
        if self.number_of_positions is not None:
            if self.number_of_positions <1:
                raise ValueError('you must analyse at least one position')

    def _validate_and_make_output(self):
        out_dir = Path(self.output_dir)
        if self.output_dir != 'Rheoscale_analysis':
            is_path = os.path.exists(self.output_dir)
            if is_path:
                pass
            else:
                print(f'the path {self.output_dir} was not found creating dir')
                os.makedirs(out_dir)
        else:
            try: 
                print(rf'attempting to create dir: {self.output_dir}\{self.protein_name}_Rheoscale')
                os.makedirs((out_dir / f'{self.protein_name}_Rheoscale'))
                
            except:
                print(rf'file: {self.output_dir}\{self.protein_name}_Rheoscale aready exists please move dir or delete or rerun with new name')    

    def change_colums(self, **updates: str) -> "RheoscaleConfig":
        new_cols = dict(self.columns)
        for key, value in updates.items():
            if key not in new_cols:
                raise KeyError(f"Invalid column key: {key}")
            new_cols[key] = value
        return replace(self, columns=MappingProxyType(new_cols))
    
    def numeric_or_none_dict(self) -> NumericFieldDict:
        """
        Returns only fields whose declared type includes int or float,
        with values restricted to int | float | None.
        """
        out: dict[str, None | Number] = {}

        for f in fields(self):
            val = getattr(self, f.name)

            origin = get_origin(f.type)
            args = get_args(f.type)

            if val is None and Number in args:
                out[f.name] = None
                continue

            # Exclude bool explicitly
            if isinstance(val, bool):
                continue

            
            if f.type in (int, float):
                out[f.name] = val
            elif origin is Union and any(t in (int, float) for t in args):
                out[f.name] = val

        return FixedKeysDict(out)


    def validate(self):
        if self.requires_input_file and not self.input_file_name:
            raise ValueError(
                "Config requires an input file, but input_file_name is not set."
            )

        if self.Min_Override and self.Min_Override_Val is None:
            raise ValueError("Min_Override is True but Min_Override_Val is None")

        if self.Max_Override and self.Max_Override_Val is None:
            raise ValueError("Max_Override is True but Max_Override_Val is None")

        if self.Error_Override and self.Error_Override_Val is None:
            raise ValueError("Error_Override is True but Error_Override_Val is None")

    def to_json(self, path: str):
        import json
        with open(path, "w") as f:
            json.dump(asdict(self), f, indent=4)

    @classmethod
    def from_json(cls, path: str):
        import json
        with open(path) as f:
            return cls(**json.load(f))


# io.py
import pandas as pd, os
from types import MappingProxyType
from typing import Optional
from pathlib import Path
from .config import RheoscaleConfig
from .plotting_histograms import plot_all_positions

def load_data(config: RheoscaleConfig) -> pd.DataFrame:

    # Otherwise load from file
    if config.input_file_name is not None:
        DMS_data = pd.read_csv(config.input_file_name)
        
        return DMS_data
    else:
        raise ValueError('Either a input file must be entered within the configeration XOR a df must be added during this call')
    
    
def validate_columns(df: pd.DataFrame, config: RheoscaleConfig):
    missing = set(config.columns.values()) - set(df.columns)

    if not missing:
        print("Column names in config match the CSV or DataFrame")
    else:
        raise ValueError(
            f"Missing columns in DataFrame: {missing}"
        )
    
def write_outputs(running_config: RheoscaleConfig, position_df: pd.DataFrame):
    out_dir = (Path(running_config.output_dir) / f'{running_config.protein_name}_Rheoscale')
    
    json_path = out_dir / f'{running_config.protein_name}_running_config.json'
    #make_config_output 
    print('hello')
    running_config.to_json(json_path)

    #make data sheet output 
    data_path = out_dir / f'{running_config.protein_name}_raw_data.csv'
    position_df.to_csv(data_path)

    #just classifcations
    just_pos_and_assign = position_df[['position', 'assignment']]
    class_path = out_dir / f'{running_config.protein_name}_classifications.csv'
    just_pos_and_assign.to_csv(class_path)

    if running_config.output_histogram_plots:
        is_hist_plots = True
    else:
        is_hist_plots= False
    if running_config.dead_extremum == "Min":
        dead_value = running_config.min_val
    else:
        dead_value = running_config.max_val

    position_list = position_df['position'].to_list()
    hist_list = position_df['histogram'].to_list()
    plot_output = out_dir / f'{running_config.protein_name}_plots'
    os.makedirs(plot_output, exist_ok=True)
    print(running_config.WT_val)
    plot_all_positions(position_list, hist_list, running_config.dead_extremum, running_config.WT_val,dead_value ,plot_output, running_config.neutral_binsize,running_config.protein_name, is_hist_plots, is_even_bins=running_config.even_bins)




